from jinja2 import Environment, PackageLoader, select_autoescape
from mprov_jobserver.plugins.plugin import JobServerPlugin
import os


jenv = Environment(
    loader=PackageLoader("mprov_jobserver"),
    autoescape=select_autoescape()
)

class DnsmasqDNSConfig(JobServerPlugin):
    dnsmasqConfDir=''
    mprovDnsmasqDir=''
    def load_config(self):
        return True
    def handle_jobs(self):

        # get the network informatioin
        response = self.js.session.get( self.js.mprovURL + 'networks/?managedns=True')
        data = {
            'networks': response.json(),
            'enableDNS': True,
        }
        with open(self.dnsmasqConfDir + 'dns.conf', 'w') as conf:
            conf.write(jenv.get_template('dnsmasq/dns.conf.j2').render(data))

        # make sure the mprov hosts dir exists
        os.makedirs(self.mprovDnsmasqDir + '/dns/', exist_ok=True)
        # now the interfaces
        for network in data['networks']: 
        
            # grab the network interface.
            response = self.js.session.get( self.js.mprovURL + 'networkinterfaces/?network=' + network['slug'])

            data_hosts = {
                'enableDNS': True,
                'hosts': response.json(),
            }

            # merge in the switches
            response = self.js.session.get( self.js.mprovURL + 'switches/?network=' + network['slug'])
            data_hosts['hosts'] = data_hosts['hosts'] + (response.json())
            # print(data_hosts)
            for idx, host in enumerate(data_hosts['hosts']):
                if 'mgmt_mac' in host:
                    data_hosts['hosts'][idx]['mac'] = host['mgmt_mac']
                if 'mgmt_ip' in host:
                    data_hosts['hosts'][idx]['ipaddress'] = host['mgmt_ip']
                data_hosts['hosts'][idx]['domain'] = network['domain']
                # print(data_hosts['hosts'][idx])

            with open(self.mprovDnsmasqDir + '/dns/' + network['slug'] + '-dns.conf', 'w') as conf:
                conf.write(jenv.get_template('dnsmasq/dns_host.conf.j2').render(data_hosts))




from .dnsmasq_mod.config import DnsmasqConfig
from .dnsmasq_mod.dns import DnsmasqDNSConfig
from .dnsmasq_mod.dhcp import DnsmasqDHCPConfig
from .plugin import JobServerPlugin

# class to handle dnsmaq generation.
# This class can job modules:
# - pxe-update
# - pxe-delete
# - dns-update
# - dns-delete
# - dhcp-update
# - dhcp-delete
# If other jobmodules are running that can consume these jobmodules,
# you may have unexpected results.
class dnsmasq(JobServerPlugin):
  jobModule = 'dnsmasq'
  enableDNS=False
  enableDHCP=False
  enableTFTP=False
  dnsmasqConfDir=''
  mprovDnsmasqDir=''
  tftproot=''
  undionlyImg = ''
  dnsmasqUser=''
  threads = []
  def handle_jobs(self):
    # we can also run the DNS thread
    if self.enableDNS:
      # grab any DNS jobs.
      # See if we have any image-delete jobs, and take 'em if we do, else just exit
      jobquery = "&module=[\"dns-update\",\"dns-delete\"]"
      # print(jobquery)
      if not self.js.update_job_status(self.jobModule, 2, jobquery=jobquery + "&status=1"):
        pass # no jobs.
        self.enableDNS = False
      else:
        dnsThread = DnsmasqDNSConfig(self.js)
        dnsThread.dnsmasqConfDir = self.dnsmasqConfDir
        dnsThread.mprovDnsmasqDir = self.mprovDnsmasqDir
        dnsThread.start()
        self.threads.append(dnsThread)

    # and finally the DHCP thread
    if self.enableDHCP:
      # grab any DHCP/PXE jobs.
      # See if we have any image-delete jobs, and take 'em if we do, else just exit
      jobquery = "&module=[\"pxe-update\",\"dhcp-update\",\"pxe-delete\",\"dhcp-delete\"]"
      # print(jobquery)
      if not self.js.update_job_status(self.jobModule, 2, jobquery=jobquery + "&status=1"):
        pass # no jobs.
        self.enableDHCP = False
      else:
        dhcpThread = DnsmasqDHCPConfig(self.js)
        dhcpThread.dnsmasqConfDir = self.dnsmasqConfDir
        dhcpThread.mprovDnsmasqDir = self.mprovDnsmasqDir
        dhcpThread.tftproot=self.tftproot
        
        dhcpThread.enableTFTP = self.enableTFTP
        dhcpThread.start()
        self.threads.append(dhcpThread)
    # Based on our settings, let's start up the submodules for dnsmasq.
    if(self.enableDNS or self.enableDHCP):
      
      # DNS or DHCP is on, so let's run the config module
      confThread = DnsmasqConfig(self.js)
      confThread.dnsmasqConfDir=self.dnsmasqConfDir
      confThread.mprovDnsmasqDir = self.mprovDnsmasqDir
      confThread.tftproot = self.tftproot
      confThread.undionlyImg = self.undionlyImg
      confThread.dnsmasqUser = self.dnsmasqUser
      confThread.start()
      self.threads.append(confThread)    

    for thread in self.threads:
      thread.join()
      



from datetime import datetime
import yaml
import time
import requests
import sys
import socket
import platform
import json
import os
from inspect import isclass
from .plugins.plugin import JobServerPlugin
import mprov_jobserver.plugins
import glob



class JobServer ():
  mprovURL = "http://127.0.0.1:8080/"
  running = True
  session = requests.Session()
  job_module_plugins = {}
  heartbeatInterval = 10
  configfile="/etc/mprov/jobserver.yaml"
  ip_address = ""
  myaddress = ""
  plugin_dir = os.path.dirname(os.path.realpath(__file__)) + '/plugins/'
  jobmodules = []
  running_threads = {}
  config_data = {}
  apikey = ""
  sessionOk = False
  id = -1
  runonce = False
  register = True


  def __init__(self, **kwargs):
    print("mProv Job Server Starting.")


    # if we get passed in a config file, let's use that
    if 'configfile' in kwargs:
      self.configfile = kwargs['configfile']
    
    # load our config
    self.load_config()
    
    
    # set runonce if someone sent it to us.
    if 'runonce' in kwargs:
      self.runonce = kwargs['runonce']
      
    if '-d' in set(sys.argv):
      self.register = False
    # Load Plugins. (plugins register job modules.)
    self.load_plugins()

    # Authenticat to the Control Center and start a session
    print("mProv Job Server authenticating.")
    if not self.startSession():
        print("Error: Unable to log into mProv Control Center.",file=sys.stderr)

    # register the server.
    self.register_server()

    pass

  def yaml_include(self, loader, node):
      # Get the path out of the yaml file
    file_name = os.path.join(os.path.dirname(loader.name), node.value)
    
    # we have a glob, so we will iterate.
    result = {}
    for file in glob.glob(file_name):
      with open(file) as inputfile:
        result.update(yaml.load(inputfile, Loader=yaml.FullLoader)[0])
    return result



  def load_config(self):
    # load the config yaml
    # print(self.configfile)
    yaml.add_constructor("!include", self.yaml_include)
    if not(os.path.isfile(self.configfile) and os.access(self.configfile, os.R_OK)):
      self.configfile = os.getcwd() + "/jobserver.yaml"
    # print(self.configfile)
    if not(os.path.isfile(self.configfile) and os.access(self.configfile, os.R_OK)):
      print("Error: Unable to find a working config file.")
      print("Try passing one in with the '-c' option.")
      sys.exit(1)


    with open(self.configfile, "r") as yamlfile:
      self.config_data = yaml.load(yamlfile, Loader=yaml.FullLoader)

    # flatten the config space
    result = {}
    for entry in self.config_data:
      result.update(entry)
    self.config_data = result
  
    # pp = pprint.PrettyPrinter(indent=2,width=100,)
    # pp.pprint(self.config_data)
    # map the global config on to our object
    for config_entry in self.config_data['global'].keys():
      try:
        getattr(self, config_entry)
        setattr(self, config_entry, self.config_data['global'][config_entry])
      except:
        print("Error: " + config_entry + " is not a valid config entry in 'global'.", file=sys.stderr)
        sys.exit(1)
    pass

  def load_plugins(self):
    if self.jobmodules is None:
      print("Error: You must specify at least 1 jobmodule!")
      sys.exit(1)
    # Load plugins set in the config file.
    for mod in self.jobmodules:
      # self.job_module_plugins[mod] = importlib.import_module('.' + mod, 'mprov.mprov_jobserver.plugins')
      attribute = getattr(mprov_jobserver.plugins, mod.replace('-', '_'))
        # attribute = getattr(attribute, attribute_name.replace('-', '_'))
      if isclass(attribute) and issubclass(attribute, JobServerPlugin):
        globals()[mod.replace('-', '_')] = attribute
            
    pass

  def stop(self):
    self.running = False

  def start(self):
    # Start processing jobs.
    # this strange looking loop allows us to die quickly on SIGTERM
    counter=self.heartbeatInterval
    while(self.running):
      if self.sessionOk is False:
        self.startSession()
      else:
        if(counter == self.heartbeatInterval):
          for mod in self.jobmodules:
            # first check if the thread is done running.
            if mod in self.running_threads: 
              # check if this thread is still running
              if not self.running_threads[mod].isAlive() :
                # print ("Thread " + mod + " ended.")
                # if it's done running remove it.
                self.running_threads[mod].handled = True
                del self.running_threads[mod]

            # if a thread of this plugin is not running, start one.
            if mod not in self.running_threads:
              # print ("Starting mod... " + mod)
              mod_cls = getattr(mprov_jobserver.plugins, mod.replace('-', '_'))
              mod_cls = getattr(mod_cls, mod.replace('-', '_'))
              self.running_threads[mod] = mod_cls(self)
              self.running_threads[mod].start()
              
          #print(".", sep=None)
          if not self.runonce:
            self.register_server()
          counter=0
        counter+=1
        time.sleep(1)
        if self.runonce:
          print("Job Server in 'runonce' mode.")
          self.running = False
    print("\nJob Server Exiting...")
    # wait for any jobmodules to complete.
    for mod in self.jobmodules:
        # first check if the thread is done running.
        if mod in self.running_threads: 
          # check if this thread is still running
          # print("\tWaiting on thread " + mod)
          self.running_threads[mod].join()
          
    
    return 0


  
  def update_job_status(self, job_module, status, jobid=None,jobquery=""):
    data = {
      'pk': job_module,
      'status': status,
    }
    if(status == 2) :
      # setting the job to running
      data['start_time'] = datetime.now().strftime("%Y-%m-%dT%H:%M:%S")
      data['end_time'] = None
    elif(status == 3) or (status == 4) :
      # job failed
      data['end_time'] = datetime.now().strftime("%Y-%m-%dT%H:%M:%S")
    if jobid is not None:
      # get the specific job status
      queryURL = self.mprovURL + 'jobs/' + str(jobid) + '/'
    elif jobquery != "":
      queryURL = self.mprovURL + 'jobs/?' + jobquery
    else:
      queryURL = self.mprovURL + 'jobs/?search=' + job_module
    response = self.session.get( queryURL )
    if response.status_code == 400:
        print("Error: Server returned error 400. Make sure your specified jobmodules exist.",file=sys.stderr)
        exit(1)
    if (response.json() == [] ):
      return False
    # print(job_module)
    # print(response.json())
    updateCount = 0
    jobs = response.json()
    # single objects still need to be in a list.
    if type(jobs) != list:
      jobs = [jobs]
    for job in jobs:

      # no need to update if our status hasn't changed.
      # print("\t" + str(job['id']))
      if(job['status'] == status ):
        continue

      # if we are in an end state, don't update.
      if(job['status'] == 3) or (job['status'] == 4):
        continue

      data['id'] = job['id']
      data['jobserver'] = self.id
      # print(data)
      response = self.session.patch(self.mprovURL + 'jobs/' + str(data['id']) + '/', data=json.dumps(data))
      updateCount += 1

      if response.status_code == 400:
        print("Oops!, JobID: " + str(data['id']))
        print(response.json())
        continue

    return updateCount
    # return True

    #print(job_module + " " + str(status))
    pass

  def register_server(self):
    if not self.register:
      return
    # get my hostname from platform
    myHostname = platform.node()

    # setup or server info register payload
    data = {
        'name': myHostname,
        'address': self.ip_address,
        'jobmodules': self.jobmodules,

    }
    # print(self.config_data)
    if 'image-server' in self.config_data:
    # if  self.config_data['image-server']:
      if 'serverPort' in self.config_data['image-server']:
        # print(self.config_data['image-server']['serverPort'])
        data['port'] = self.config_data['image-server']['serverPort']
    
    #print(data)
    # post the response (maybe should be put?) to the server.
    try: 
      response = self.session.post(self.mprovURL + 'jobservers/', data=json.dumps(data), )
      
    except:
      self.sessionOk = False
      self.startSession()
      return
    
    if response.status_code == 400:
        print("Error: Server returned error 400. Make sure your specified jobmodules exist.",file=sys.stderr)
        exit(1)
    if response.status_code == 500:
        print("Error: The mPCC had an internal server error.")
        exit(1)
    # pp = pprint.PrettyPrinter(indent=2,width=100,)
    # pp.pprint(vars(response))
    # # print(response.text)
    if type(response.json()) is dict:
      print("Error: Invalid response from mPCC")
      print(response.json())
      sys.exit(1)
    # print(response.json())
    result = json.loads(response.json())
    
    # grab our id from the MPCC
    self.id = result['pk']

    print("Heartbeat - " + datetime.now().strftime("%d/%m/%Y %H:%M:%S"))


  def startSession(self):
    
    self.session.headers.update({
      'Authorization': 'Api-Key ' + self.apikey,
      'Content-Type': 'application/json'
      })

    # test connectivity
    try:
      response = self.session.get(self.mprovURL, stream=True)
    except:
      print("Error: Communication error to the server.  Retrying.", file=sys.stderr)
      self.sessionOk = False
      time.sleep(10)
      return
    self.sessionOk = True
    if self.myaddress is not None:
      if self.myaddress != '':
        self.ip_address = self.myaddress
      else:
        print("Warning: No address set in config, attempting autodetection.  This may not work right...")
        # get the sock from the session
        s = socket.fromfd(response.raw.fileno(), socket.AF_INET, socket.SOCK_STREAM)
        # get the address from the socket
        address = s.getsockname()
        self.ip_address=address
      
    # if we get a response.status_code == 200, we're ok.  If not,
    # our auth failed.
    return response.status_code == 200

 


from .plugin import JobServerPlugin
import os
import shutil

class image_delete(JobServerPlugin):
  jobModule = 'image-delete'
  imageList = None
  imageDir = ""
  def handle_jobs(self):
    # See if we have any image-delete jobs, and take 'em if we do, else just exit
    jobquery = "jobserver=" + str(self.js.id) + "&module=image-delete"
    if not self.js.update_job_status(self.jobModule, 2, jobquery=jobquery + "&status=1"):
      return # no jobs.
    
    # we have some jobs, let's do this.
    # get the jobs
    response = self.js.session.get( self.js.mprovURL + 'jobs/?' + jobquery + '&status=2')
    for job in response.json():
      # grab the params.
      try:
          params = job['params']
      except: 
        print("Error: Image Delete Job with no imageID present, cannot parse params")
        self.js.update_job_status(self.jobModule, 3, jobid=job['id'])
        return
      if self.imageList is None or params['imageId'] in self.imageList:
        # This appears to be one of ours.
        if os.path.exists(self.imageDir):
          if os.path.exists(self.imageDir + '/' + params['imageId']):
            # the path seems there.  Kill it.
            shutil.rmtree(self.imageDir + '/' + params['imageId'])
            #print("rm -rf " + self.imageDir + '/' + params['imageId'])
            self.js.update_job_status(self.jobModule, 4, jobid=job['id'])
          else:
            print("Error: image has no directory:" + self.imageDir + '/' + params['imageId'])
            self.js.update_job_status(self.jobModule, 3, jobid=job['id'])
            
        else:
          print("Error: imageDir doesn't exist: " + self.imageDir)
          self.js.update_job_status(self.jobModule, 3, jobid=job['id'])
          
      else:
        print("Error: Recieved image-delete job for image we don't host. HuH?")
        self.js.update_job_status(self.jobModule, 3, jobid=job['id'])

  


import shutil
from threading import local
from .plugin import JobServerPlugin
import os
import json
import sys

class image_sync(JobServerPlugin):
  jobModule = 'image-sync'
  imageDir = ""
  imageList = None

  # override default load config because we have to check that image-server is also enabled.
  def load_config(self):
    if 'image-server' not in self.js.jobmodules:
      print("Configuration Error: you MUST run image-server on image-sync nodes!")
      print("                   : image-sync nodes are SOURCE nodes for images!")
      print("                   : and need a way to serve images! Job Module Halted!!!")
      sys.exit(1)
    return super().load_config()

  def handle_jobs(self):
    
    # image-sync is not job based.  We will connect to the MPCC, compare the version we know
    # about for our image, and update it if it doesn't match.

    # images we care about: self.imageList, if populated.  All images in MPCC if None.
    # image versions tacked via version file in self.imageDir.

    if not os.path.exists(self.imageDir):
      # if we don't have an image path, make one.
      os.makedirs(self.imageDir)
    
    # see if we have a image-versions file
    if not os.path.exists(self.imageDir + '/image-versions'):
      # no versions file.  Create a blank one right quick
      with open(self.imageDir + '/image-versions', "w") as vfile:
        vfile.write(json.dumps({}))
    
    # now we should have a dir and a version file, lets get our list of images we care about
    if self.imageList is None:
      self.imageList = []
      # no imageList in config, grab it from MPCC
      response = self.js.session.get( self.js.mprovURL + 'images/')
      for image in response.json():
        self.imageList.append(image['slug'])
    
    # so here we should have either a hard coded config list of images, 
    # or a list of all images grabbed from the server.
    for image in self.imageList: 
      # grab the MPCC version for this image.
      # if MPCC 404's the image was deleted.  Remove from our imageList
      # print an error, and move on.  File deletion happens below.
      response = self.js.session.get( self.js.mprovURL + 'systemimages/' + image + '/')
      if(response.status_code == "404" ):
        # image not found in the MPCC, must have been removed, delete it locally.
        self.imageList.remove(image)
        print("Error: Image '" + image + "' was not found on the MPCC, removing locally." )
        print("Error: Consider removing it from your config.")
        continue
      if(response.json()['jobservers'] == []):
        self.imageList.remove(image)
        print("Warn: No jobservers available yet for image: " + image)
        print("Warn: Not syncing image: " + image)
        continue
      mpccVersion = response.json()['version']
      currJobServers = response.json()['jobservers']

      with open(self.imageDir + '/image-versions', "r") as vfile:
        # grab the version info into a var
        try:
          imgVersions = json.loads(vfile.read())
        except:
          # if we have an error loading the version data, 
          # init it with no version data, we'll rebuild it 
          # in a moment.
          imgVersions = {}
    
      # see if our image is in imgVersions
      ourVersion = 0
      if image in imgVersions:
        ourVersion = imgVersions[image]
      
      # compare our version to the one the MPCC has for us.
      if ourVersion != mpccVersion:
        # Versions don't match, ask the MPCC to grab a new copy.
        with self.js.session.get(self.js.mprovURL + 'images/' + image, stream=True) as remoteImage:
          remoteImage.raise_for_status()
          os.makedirs(self.imageDir + '/' + image,exist_ok=True)
          with open(self.imageDir + '/' + image + '/' + image + '.img', 'wb') as localFile:
            for chunk in remoteImage.iter_content(chunk_size=8192):
              localFile.write(chunk)

          # grab the initramfs
          with self.js.session.get(self.js.mprovURL + 'images/' + image + '.initramfs', stream=True) as remoteImage:
            remoteImage.raise_for_status()
            os.makedirs(self.imageDir + '/' + image,exist_ok=True)
            with open(self.imageDir + '/' + image + '/' + image + '.initramfs', 'wb') as localFile:
              for chunk in remoteImage.iter_content(chunk_size=8192):
                localFile.write(chunk)
        
          # grab the kernel
          with self.js.session.get(self.js.mprovURL + 'kernels/' + image + '.vmlinuz', stream=True) as remoteImage:
            remoteImage.raise_for_status()
            os.makedirs(self.imageDir + '/' + image,exist_ok=True)
            with open(self.imageDir + '/' + image + '/' + image + '.vmlinuz', 'wb') as localFile:
              for chunk in remoteImage.iter_content(chunk_size=8192):
                localFile.write(chunk)
                
        # file download complete.  Update our version
        imgVersions[image] = mpccVersion
        ourVersion = mpccVersion
        with open(self.imageDir + '/image-versions', "w") as vfile:
          vfile.write(json.dumps(imgVersions))

        # tell the MPCC we can host this file
        jobservers = []
        for jobserver in currJobServers:
          jobservers.append(jobserver['id'])
        # now append our id
        jobservers.append(self.js.id)
        data = {
          'slug': image,
          'needs_rebuild': False,
          'jobservers': jobservers,
        }
        response = self.js.session.patch(self.js.mprovURL + 'systemimages/' + str(data['slug']) + '/', data=json.dumps(data))

    # Clean up any directories that are not in our imageList.
    # print("Scanning for haning images... " + self.imageDir)
    for entry in os.listdir(self.imageDir):
      # print(entry)
      if os.path.isdir(self.imageDir + '/' + entry):
        # see if this entry is in our image list
        if not os.path.basename(entry) in self.imageList:
          # Nope doesn't seem to be there, doesn't need to exist on disk then.
          #print("rm -rf " + self.imageDir + '/' + entry)
          print("Warn: Removed unknown image directory: " + self.imageDir + '/' + entry)
          shutil.rmtree(self.imageDir + '/' + entry)

        



from .plugin import JobServerPlugin

class pxe_update(JobServerPlugin):
  jobModule = 'pxe-update'
  tftpPath = "/tftpboot/"
  networks = None

  def handle_jobs(self):
    # See if we have any pxe-update jobs, and take 'em if we do, else just exit
    jobquery = "jobserver=" + str(self.js.id) + "&module=pxe-update"
    if not self.js.update_job_status(self.jobModule, 2, jobquery=jobquery + "&status=1"):
      return # no jobs.
    
    # we have some jobs, let's do this.
    # get the jobs
    response = self.js.session.get( self.js.mprovURL + 'jobs/?' + jobquery + '&status=2')
    for job in response.json():


      # if networks is None, grab all the networks.  
      if self.networks is None:
        self.networks = []
        # no imageList in config, grab it from MPCC
        response = self.js.session.get( self.js.mprovURL + 'networks/')
        for network in response.json():
          self.networks.append(network['slug'])
      
      # so here we should have either a hard coded config list of networks, 
      # or a list of all networks grabbed from the MPCC.
      for network in self.networks: 
        
        # grab the network interface.
        response = self.js.session.get( self.js.mprovURL + 'networkinterfaces/?network=' + network)

        response = self.js.session.get( self.js.mprovURL + 'networkinterfaces/' + network + '/details')


#!/usr/bin/python
import importlib.resources
import os
import signal
from mprov_jobserver.app import JobServer
import sys

jobServer = None

# define our exit handler.
def exitHandler(signum, frame):
    if jobServer is not None:
        jobServer.stop()
        
    else:
        print("jobServer is None?")
        sys.exit(1)

signal.signal(signal.SIGINT, exitHandler)
signal.signal(signal.SIGTERM, exitHandler)

def main():
    global jobServer
    runonce=False

    if '-r' in set(sys.argv):
        runonce = True
    
    if '-c' in set(sys.argv):
        configfile = sys.argv[sys.argv.index('-c') + 1]
    else:
        configfile = "/etc/mprov/jobserver.yaml"
        # using the default config file, see if one exists.
        if not os.path.isfile(configfile):
            # no config file in the default location, let's copy one in and 
            # say something to the user about editing it.
            for entry in importlib.resources.contents('mprov_jobserver'): 
                if entry[-4:] == "yaml":
                    print(entry)

                    # make the standard config directory
                    os.makedirs('/etc/mprov/', exist_ok=True)
                    try: 
                        with importlib.resources.open_text('mprov_jobserver', entry) as yaml_file:
                            with open('/etc/mprov/' + entry, "w") as yaml_file_out:
                                yaml_file_out.write(yaml_file.read())
                    except:
                        print("Error copying " + entry + " to /etc/mprov/")
                        sys.exit(1)
                    
            # that succeeded so let's grab the plugin information
            os.makedirs('/etc/mprov/plugins', exist_ok=True)
            for p_entry in importlib.resources.contents('mprov_jobserver.plugins'):
                if p_entry[-4:] == 'yaml':
                    try:
                        with importlib.resources.open_text('mprov_jobserver.plugins', p_entry) as p_yaml_file:
                            with open('/etc/mprov/plugins/' + p_entry, 'w') as p_yaml_out_file:
                                p_yaml_out_file.write(p_yaml_file.read())
                    except:
                        print('Error copying ' + p_entry + ' to /etc/mprov/plugins/')
                        sys.exit(1)
            
            # if we make it here, we have copied the config to /etc/mprov/ so prompt the user to go modify it
            print('A fresh config has been copied to /etc/mprov/  Please go there')
            print('and edit the .yaml files in that directory to setup your environment.')
            sys.exit(1)
            
    jobServer = JobServer(runonce=runonce, configfile=configfile)
    
    if jobServer is not None:
        # Start the main loop and run the plugin handling routines.
        return jobServer.start()
    return 1
def __main__():
    main()
if __name__ == "__main__" or __name__ == 'mprov.mprov_jobserver':
    main()



import time
from .plugin import JobServerPlugin
from time import sleep
import os
import sys
import json
from urllib.parse import urlparse
import yaml
from jinja2 import Environment, PackageLoader, select_autoescape

jenv = Environment(
    loader=PackageLoader("mprov_jobserver"),
    autoescape=select_autoescape()
)



class image_update(JobServerPlugin):
  jobModule = 'image-update'
  imageDir = ""
  imageList = None


# rpm2cpio < rocky-repos-8.5-3.el8.noarch.rpm | cpio -D /mnt/tmproot/ -id
# dnf -y --installroot=/mnt/tmproot/ --releasever=8 groupinstall 'Minimal Install'

  # override default load config because we have to check that image-server is also enabled.
  def load_config(self):
    if 'image-server' not in self.js.jobmodules:
      print("Configuration Error: you MUST run image-server on image-upate/image-delete nodes!")
      print("                   : image-update/delete nodes are SOURCE nodes for images!")
      print("                   : and need a way to serve images! Job Module Halted!!!")
      sys.exit(1)
    return super().load_config()

  def handle_jobs(self):
    if self.imageList is None:
      # grab all the image-update jobs.
      if not self.js.update_job_status(self.jobModule, 2):
        return # no jobs.
      # populate self.imageList with the list of all pending images.
      self.imageList = []
      response = self.js.session.get( self.js.mprovURL + 'jobs/?jobserver=' + str(self.js.id) + '&module=' + self.jobModule + '&status=2')
      for job in response.json():
        try:
          params = job['params']
        except: 
          print("Error: Image Update Job with no imageID present, cannot parse params")
          self.js.update_job_status(self.jobModule, 3, jobid=job['id'])
          return
        try:
          self.imageList.append(params['imageId'])
        except:
          print("Error: Image Update Job failed, corrupted params.")
          self.js.update_job_status(self.jobModule, 3, jobid=job['id'])
          return
    else:
      # imageList is not none, so let's grab just the ones we care about.
      updateFound = False
      for image in self.imageList:
        # json to pass to the job search.
        paramstr="{\"imageId\":\"" + image + "\"}"
        if self.js.update_job_status(self.jobModule, 2, jobquery=paramstr):
          updateFound=True
        else: 
          # this image wasn't requested to be updated, remove it from the list.
          self.imageList.remove(image)
      if (not updateFound) or len(self.imageList) == 0:
        return # no updates found.
        

    # check if the imageDir exists.
    if not os.path.exists(self.imageDir):
      # if not, attempt to make it, recursively.
      try:
        os.makedirs(self.imageDir, exist_ok=True)
      except:
        print("Error: Unable to make imageDir: " + self.imageDir)
        self.js.update_job_status(self.jobModule, 3, jobquery='jobserver=' + str(self.js.id) + '&status=2')
        return
    
    # now we iterate over our imageList, and build the image(s)
    for image in self.imageList:
      print(f"Processing update for image {image}")
      query = 'systemimages/' + image + '/'
      response = self.js.session.get( self.js.mprovURL + query)
      if response.status_code == 200:
        imageDetails = response.json()
        baseURL=imageDetails['osdistro']['baserepo']['repo_package_url']
        print(baseURL)

        imgDir = self.imageDir + '/' + imageDetails['slug']
        # create this image's dir.  Use the image['slug']
        os.makedirs(imgDir, exist_ok=True)
        os.chdir(imgDir)
        # grab the repo rpm
        url = urlparse(baseURL)
        file = os.path.basename(url.path)
        print("Grabbing os repo package: " + baseURL)
        if os.system('wget -O ' + file + ' ' + baseURL ):
          print("Error: unable to get repo package: " + baseURL)
          self.js.update_job_status(self.jobModule, 3, jobquery='jobserver=' + str(self.js.id) + '&status=2')
          return

        # force the RPM to unpack to our image dir.
        print("Unpacking RPM to " + imgDir)
        
        if os.system('rpm2cpio < ' + file + ' | cpio -D ' + imgDir + ' -id'):
          print("Error: unable to extract repo package: " + file + ' into ' + imgDir)
          self.js.update_job_status(self.jobModule, 3, jobquery='jobserver=' + str(self.js.id) + '&status=2')
          return


        # build the filesystem.
        if os.system('dnf -y --installroot=' + imgDir + ' --releasever=' + str(imageDetails['osdistro']['version'])  + ' groupinstall \'Minimal Install\''):
          print("Error: unable to genergate image filesystem.")
          self.js.update_job_status(self.jobModule, 3, jobquery='jobserver=' + str(self.js.id) + '&status=2')
          return

        # install and copy the kernel image to the image root
        if os.system('dnf -y --installroot=' + imgDir + ' --releasever=' + str(imageDetails['osdistro']['version'])  + ' install kernel python38 python38-pyyaml python38-requests python38-jinja2.noarch jq'):
          print("Error uanble to install kernel into image filesystem")
          self.js.update_job_status(self.jobModule, 3, jobquery='jobserver=' + str(self.js.id) + '&status=2')
          return

        if os.system('ls -t ' + imgDir + '/boot/vmlinuz-* | grep -v rescue | head -n1 | xargs -I{} cp {} ' + imgDir + '/' + image + '.vmlinuz'): 
          print("Error: unable to copy kernel image.")
          self.js.update_job_status(self.jobModule, 3, jobquery='jobserver=' + str(self.js.id) + '&status=2')
          return

        if os.system('ls -t ' + imgDir + '/boot/initramfs-* | grep -v rescue | head -n1 | xargs -I{} cp {} ' + imgDir + '/' + image + '.initramfs'): 
          print("Error: unable to copy initramfs.")
          self.js.update_job_status(self.jobModule, 3, jobquery='jobserver=' + str(self.js.id) + '&status=2')
          return

        if os.system('rm -f ' + imgDir + '/dev/ram0; mknod -m 600 ' + imgDir + '/dev/ram0 b 1 0'):
          print('Error: trying to make /dev/ram0')
          self.js.update_job_status(self.jobModule, 3, jobquery='jobserver=' + str(self.js.id) + '&status=2')
          return
        
        if os.system('rm -f ' + imgDir + '/dev/initrd; mknod -m 400 ' + imgDir + '/dev/initrd b 1 250'):
          print('Error: trying to make /dev/ram0')
          self.js.update_job_status(self.jobModule, 3, jobquery='jobserver=' + str(self.js.id) + '&status=2')
          return
          
        
        # run image-gen scripts.
        
        # grab a copy of the jobserver wheel from the main mprov server
        # TODO: Change this to a pip install mprov_jobserver command after publication

        if os.system("chroot " + imgDir + " pip3 --no-cache-dir install mprov_jobserver --force-reinstall"):
          print("Error: Unable to install mprov_jobserver python module.")
          return

        # check if the imagDir + /tmp/mprov exists and create it if not
        os.makedirs(imgDir + '/tmp/mprov/plugins/', exist_ok=True)
        
        # if os.system('rsync -ar ' + str(scriptDir) + '/../../mprov_jobserver ' + imgDir + '/root/mprov/'):
        #   print("Error: Unable to copy a jobserver to the image.")
        #   self.js.update_job_status(self.jobModule, 3, jobquery='jobserver=' + str(self.js.id) + '&status=2')
        #   return
        
        # create a config file to /tmp for our script-runner instance.
        localConfig = [dict()]
        localConfig[0]['global'] = self.js.config_data['global']
        print(localConfig[0]['global']['jobmodules'])
        localConfig[0]['global']['jobmodules'] = ['script-runner']
        localConfig[0]['global']['runonce'] = True
        # print(imgDir + '/tmp/mprov/jobserver.yaml')
        with open(imgDir + '/tmp/mprov/jobserver.yaml',"w") as confFile:
          yaml.dump(localConfig, confFile)
          confFile.write("\n- !include plugins/*.yaml")

        data = {
          'imgDir': imgDir,
          'image': image,
        }
        # now let's run our script-runner shell script.
        with open(os.open(imgDir + '/tmp/mprov/script-runner.sh',os.O_CREAT | os.O_WRONLY, 0o755) , 'w') as conf:
            conf.write(jenv.get_template('image-update/script-runner.sh').render(data))
        # now let's run our script-runner shell script.
        with open(os.open(imgDir + '/tmp/mprov/plugins/script-runner.yaml',os.O_CREAT | os.O_WRONLY, 0o755) , 'w') as conf:
            conf.write(jenv.get_template('image-update/script-runner.yaml.j2').render(data))

        if os.system(imgDir + '/tmp/mprov/script-runner.sh'):
          print("Error while running image-gen scripts via job server..")
        

        # package the filesystem into an initramfs
        print("Building " + os.getcwd() + "/" + imageDetails['slug'] + '.img')
        startTime=time.time()
        os.system('rm -f ' + imgDir + '/' + imageDetails['slug'] + '.img')
        if os.system('find .  -depth -print| cpio -H newc --quiet -oD ' + imgDir + '  | gzip -1 -c > /tmp/' + imageDetails['slug'] + '.img'):
          print("Error: unable to create initramfs")
          self.js.update_job_status(self.jobModule, 3, jobquery='jobserver=' + str(self.js.id) + '&status=2')
          return
        endTime = time.time()
        lapsed = endTime - startTime
        print("Image generated in " + str(lapsed) + " seconds.")
        print("Image saved to "+ os.getcwd() + "/" + imageDetails['slug'] + '.img')
        os.system('mv /tmp/' + imageDetails['slug'] + '.img ' + imgDir + '/' )


        # update the 'jobservers' field to be us, so that the 
        data = {
          'slug': imageDetails['slug'],
          'needs_rebuild': False,
          'jobservers':[
            self.js.id,
          ]
        }
        response = self.js.session.patch(self.js.mprovURL + 'systemimages/' + str(data['slug']) + '/', data=json.dumps(data))

        # Update our jobs with success or failure
        self.js.update_job_status(self.jobModule, 4, jobquery='jobserver=' + str(self.js.id) + '&status=2')
      else: 
        print(f"Error handling job for {image} server returned {response.status_code} for request {self.js.mprovURL + query}")
        self.js.update_job_status(self.jobModule, 3, jobquery='jobserver=' + str(self.js.id) + '&status=2')



from time import sleep
import threading
import sys

   
class JobServerPlugin(threading.Thread):
  jobModule = ""
  js = None


  def set_job_running(self):
    # This function can be overridden, especially if you are processing specific jobID's.
    return self.js.update_job_status(self.jobModule, 2) # RUNNING = 2
  
  def set_job_success(self):
    # This function can be overridden, especially if you are processing specific jobID's.
    return self.js.update_job_status(self.jobModule, 4) # SUCCESS = 4

  def set_job_failure(self):
    # This function can be overridden, especially if you are processing specific jobID's.
    return self.js.update_job_status(self.jobModule, 3) # FAILURE = 3

  def __init__(self, js):
    threading.Thread.__init__(self)
    self.name = self.__class__
    self.js = js
    pass

  def run(self):
    if self.load_config():
      self.handle_jobs()
    

  def load_config(self):
    # NOTE: Override with your config loading routine if needed. 
    # Access to js.config_data to parse the config file.
    # for information on how to parse the config_data structure, 
    # see https://pyyaml.org/wiki/PyYAMLDocumentation 
    # 
    # This data structure is loaded by the JobServer class at 
    # startup

    # This template config loader should work for all modules.  Make sure 
    # you put your module config in the plugins/ dir in the same location
    # as the main jobserver.yaml file.  
    #print(self.js.config_data)
    if self.js.config_data is None:
          print("Conf is empty? ")
          return False
    if self.jobModule not in self.js.config_data:
      # This is not necessarily an error, maybe one day print a warning?
      print("Warn: No config found for " + self.jobModule + " hope that's ok..")
      return False
    if self.js.config_data[self.jobModule] is None:
          print("Found empty module config?")
          return False
    for config_entry in self.js.config_data[self.jobModule].keys():
        try:
          getattr(self, config_entry)
          setattr(self, config_entry, self.js.config_data[self.jobModule][config_entry])
        except:
          print("Error: Unknown config entry " + config_entry + " in " + self.jobModule)
    return True

  def handle_jobs(self):
    # NOTE:
    # 
    # Default structure: 
    #
    # self.set_job_running()
    #
    # ... do something ...
    # 
    # if something :
    #   set_job_success()
    # else
    #   set_job_failure()
    #
    # By default, this function will only test the work flow with a 15 second delay as the 'work'

    # Update all pending jobs of oir module to running
    if(not self.set_job_running()):
      # no jobs to run, just return
      return

    
    print(self.jobModule + ": Sleeping 15 seconds to simulate work.  Override 'handle_jobs()' to do something useful.")
    sleep(15)
    
    # Update our jobs with success or failure
    self.set_job_success()





from http import server
import os
from .plugin import JobServerPlugin
from http.server import BaseHTTPRequestHandler, ThreadingHTTPServer

class HTTPImageServer(BaseHTTPRequestHandler):
  imageDir = ""
  def __init__(self):
    pass

  def __call__(self, *args, **kwargs):
    try: 
      super().__init__(*args, **kwargs)
    except:
      pass

  def do_GET(self):
      self.send_response(200)
      self.send_header("Content-type", "application/octet-stream")
      
      # sanitize self.path for directory escape characters.  Is this enough?
      self.path = "/" + os.path.relpath(os.path.normpath(os.path.join("/", self.path)), "/")

      #imageName = self.path.split('.', 1)[0]
      imageName = self.path.split('.', 1)[0]
      filename = self.imageDir + '/' + imageName + self.path
      # print(filename)
      size = os.path.getsize(filename)
      self.send_header("Content-Length", size)
      self.send_header("Content-Disposition", "inline; filename=\"" + self.path + "\"")
      self.end_headers()

      with open(filename, "rb") as imageFile:
        self.wfile.write(imageFile.read())
  
      

class image_server(JobServerPlugin):
  jobModule = 'image-server'
  hostName = "0.0.0.0"
  serverPort = 80
  serverInstance = None
  imageDir = ""
  def handle_jobs(self):
    print("Starting Image Server...")
    ReqHandler = HTTPImageServer()
    ReqHandler.imageDir = self.imageDir
    serverInstance = ThreadingHTTPServer((self.hostName, self.serverPort), ReqHandler)
    serverInstance.timeout=0.5
    # this should allow us to exit out ok.
    while(self.js.running):
      serverInstance.handle_request()
    
    print("Stopping Image Server.")
    #serverInstance.serve_forever()
    
    


from .plugin import JobServerPlugin
import sys
import os
import requests
from threading import Thread
import json
import subprocess
import yaml
import socket

class script_runner(JobServerPlugin):
  jobModule = 'script-runner'
  scriptTmpDir="/tmp/mprov-script-runner/"

  def depResolve(self, depDict):
    '''
    Returns an array of 'steps' for dependancy resolution.  Each 'step' guarantees each entry has no other dependencies.

      Parameters:
        depDict (dict):  is a dependency dictionary of arrays which are lists of the dependencies of their respective keys.

      Returns:
        result (list | None): a list of dicts, each entry in the list is a step of dependency resolution. None on error
    '''
    # this snippet is a bit of beautiful python borrowed from
    # https://stackoverflow.com/questions/5287516/dependencies-tree-implementation#5288547
    # and modified for clarity and implemention 

    # convert the dict of lists to a dict of dicts.
    dep_dict=dict((key, set(depDict[key])) for key in depDict)
    result=[]
    firstPass=True
    prevT=set([])
    while dep_dict:
        # print("-----")
        # print("dep_dict: " + str(dep_dict))

        # Should values not in keys produce an error? Unresolvable deps?
        # values not in keys (items without dep)
        # turns the values of the incoming dict into a set.  Then removes all the
        # keys that exist in the dict, giving us dep free items in t
        t=set(item for vals in dep_dict.values() for item in vals)-set(dep_dict.keys())
        if len(t) > 0:
          print("Error: Unresolvable Dependancies, make sure it's listed in distro, group, or entity scripts: ")
          for ent in t:
            print("\t '" + ent + "'")
          return None
        # print("t: " + str(t))
        
        # add any keys without value (items without dep) to t
        t.update(key for key, val in dep_dict.items() if not val)
        if prevT == t:
        # if not firstPass and len(t) > 0:
          # possible cyclic dependency?
          print("Error: Possibe cyclic dependency.")
          print("depDict: " + str(depDict))
          print("t: " + str(t))
          return None
        # print("t: " +  str(t))
        prevT = t
        # After the first pass, we want to check for cyclic deps.
        firstPass=False
        # t contains our dep free items for this iteration, so
        # append t to the result list.
        result.append(t)
        # print("result: " + str(result))
        
        # and cleaned up
        # for each entry in the dict, if it has a value, and we have resolved it's
        # dependancies, remove the resolved dependancies from the value
        # cleans things up in 2 ways:
        #   - removes keys that do not have a value (resolved deps)
        #   - removes values for already resolved deps from left over deps for each item.
        dep_dict=dict(((key, val-t) for key, val in dep_dict.items() if val))
        # print("dep_dict: " + str(dep_dict))
    return result

  def handle_jobs(self):
    
    # see if we are being called from a runonce command
    if not self.js.runonce:
      print("Error: script-runner must be run in a 'runonce' jobserver session.")
      return False
    print("Jobserver Running script-runner...")
    sysimage=False
    system=False
    scriptMode='image-gen' # should be img-gen or post-boot, def: img-gen
    entityId=None 

    # now we need to look for some commandline options
    if '-h' in sys.argv:
      # someone passed the -h flag, print the help.
      self.printHelp()
      return False

    # are -i and -s passed together?
    if '-i' in sys.argv and '-s' in sys.argv:
      print("Error: -i and -s are mutually exclusive!")
      self.printHelp()
      return False

    # look for a s
    if '-i' in sys.argv:
      sysimage = True
      system = False
      entityId = sys.argv[sys.argv.index('-i')+1]
    elif '-s' in sys.argv:
      sysimage = False
      system = True
      entityId = socket.gethostname()
    else:
      print("Error: You must specify -i or -s")
      self.printHelp()
      return False

    if '-b' in sys.argv:
      # someone is asking for post-boot scripts
      scriptMode = 'post-boot'
    
    # grab the scripts for this entity.
    query=""
    if sysimage:
      query="systemimages/" + entityId + "/"
    else:
      query="systems/?hostname=" + entityId 
    # print(self.js.mprovURL + query)
    response = self.js.session.get( self.js.mprovURL + query)
    # merge the scripts from distro -> system_groups -> entity
    try:
      entity = response.json()
    except:
      # if this response was bad, try to grab the nads image 
      query="systemimages/nads/"
      response = self.js.session.get( self.js.mprovURL + query )
      try:
        entity = response.json()
      except: 
        print("Error: Unable to get information from mPCC.")
        exit(1)

    # we should iterate in all the scripts.  Let's process that into a dependancy tree.
    scriptDeps = {}
    scripts = {}
    # osdistro scripts first
    for script in entity['osdistro']['scripts']:
      deps=[]
      if script['scriptType']['slug'] != scriptMode:
        continue
      if len(script['dependsOn']) > 0:
        # we have deps.
        for dep in script['dependsOn']:
          deps.append(dep['slug'])
      scriptDeps[script['slug']] = deps
      scripts[script['slug']] = script

    # then group scripts
    for group in entity['systemgroups']:
      for script in group['scripts']:
        deps=[]
        if script['scriptType']['slug'] != scriptMode:
          continue
        if len(script['dependsOn']) > 0:
          # we have deps.
          for dep in script['dependsOn']:
            deps.append(dep['slug'])
        scriptDeps[script['slug']] = deps
        scripts[script['slug']] = script

    # finally lay over the entity scirpts.
    for script in entity['scripts']:
      deps=[]
      if script['scriptType']['slug'] != scriptMode:
        continue
      if len(script['dependsOn']) > 0:
        # we have deps.
        for dep in script['dependsOn']:
          deps.append(dep['slug'])
      scriptDeps[script['slug']] = deps
      scripts[script['slug']] = script

    yaml_merged={}
    # then we will flatten config_params on the host.
    # distro first
    
    # print(entity['osdistro']['config_params'])
    tmpYamlDict = yaml.safe_load(entity['osdistro']['config_params'])
    if tmpYamlDict is not None:
      try:
        if type(tmpYamlDict) is list:
          for dict in tmpYamlDict:
            yaml_merged.update(dict)
        else: 
            yaml_merged.update(tmpYamlDict)
      except:
        pass
    
    # print(yaml_merged)
    # then the system groups
    for group in entity['systemgroups']:
      # print(group['config_params'])
      tmpYamlDict = yaml.safe_load(group['config_params'])
      if tmpYamlDict is not None:
        try:
          if type(tmpYamlDict) is list:
            for dict in tmpYamlDict:
              yaml_merged.update(dict)
          else:
              yaml_merged.update(tmpYamlDict)
        except:
          pass  
    # and finally the system/image params
    # print(yaml_merged)
    # print(yaml.safe_load(entity['config_params']))
    tmpYamlDict = yaml.safe_load(entity['config_params'])
    # print(tmpYamlDict)
    if tmpYamlDict is not None:
      try:
        if type(tmpYamlDict) is list:
          for dict in tmpYamlDict:
            yaml_merged.update(dict)
        else: 
          yaml_merged.update(tmpYamlDict)
      except:
        pass
              
    # print(yaml_merged)
    entity['config_params'] = yaml_merged

    # add the mprovURL to the entity. 
    entity['mprovURL'] = self.js.mprovURL
    
    # add our apikey to the entity
    entity['apikey'] = self.js.apikey

    # print(scriptDeps)
    # resolve our dependancies
    scriptDeps=self.depResolve(scriptDeps)
    # print(scripts)

    # make sure our local tmp dir exists
    os.makedirs(self.scriptTmpDir, exist_ok=True)
    
    # output the entity to a json file
    with open(self.scriptTmpDir + "/entity.json", "w") as entFile:
      entFile.write(json.dumps(entity))
    
    for step in scriptDeps:
      threads = []
      for script in step:
        # download the script and thread it out within this step.
        # print(scripts[script]['filename'])
        t = Thread(target=self.runScript, args=(scripts[script]['filename'], ))
        t.start()
        threads.append(t)

        # wait for the threads in this step to finish.
      for t  in threads:
        t.join()
    # TODO: remove the entity file.

    print("script-runner complete.")

  def runScript(self, filename):
    # grab the file
    local_file = self.download_file(filename)

    # run the file
    subprocess.run(local_file)
    

  def download_file(self, url):
    local_filename = url.split('/')[-1]
    local_filename = self.scriptTmpDir + "/" + local_filename
    # NOTE the stream=True parameter below
    with requests.get(url, stream=True) as r:
        r.raise_for_status()
        with open(local_filename, 'wb') as f:
            for chunk in r.iter_content(chunk_size=8192): 
                # If you have chunk encoded response uncomment if
                # and set chunk_size parameter to None.
                #if chunk: 
                f.write(chunk)
        os.chmod(local_filename, 0o755)
    return local_filename

  def printHelp(self):
    print(sys.argv[0] + ":")
    print("\t-i <systemimage-ID>    - run scripts for systemimage <systemimage-ID>")
    print("\t-s                     - run scripts for system")
    print("\t                         (NOTE: NOT the fqdn!)")
    print("\t-b                     - run in post-boot mode, defaults to image-gen mode without this flag")
    return False